unit cDIBFeatures;

{-----------------------------------------------------------------------------
The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance
with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/MPL-1.1.html

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either expressed or implied. See the License for
the specific language governing rights and limitations under the License.

The Original Code is: cDIBFeatures.PAS, released August 28, 2000.

The Initial Developer of the Original Code is Peter Morris (pete@stuckindoors.com),
Portions created by Peter Morris are Copyright (C) 2000 Peter Morris.
All Rights Reserved.

Purpose of file:
Allows component subclassing at design-time.  You can apply various descendents of
TDIBFeature to any DIB component, allowing it to move at runtime, highlight when the
mouse enters, or any other custom functionality a person designs.
New features are added by calling the RegisterDIBFeature command.

Contributor(s):
None as yet


Last Modified: August 28, 2000
Current Version: 1.00

You may retrieve the latest version of this file at the Project JEDI home page,
located at http://www.delphi-jedi.org
or at http://www.stuckindoors.com/dib


Known Issues:
To be updated !
-----------------------------------------------------------------------------}


interface
uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs;

type
  TAllowFeatureEvent = procedure(Sender : TObject; Control : TControl; var Allow : Boolean) of object;

  TControlItem = class(TCollectionItem)
  private
    FControl                  : TControl;
    procedure SetControl(const Value: TControl);
  protected
  public
    procedure AssignTo(Dest : TPersistent); override;
    function GetDisplayName : String; override;
  published
    property Control          : TControl
      read FControl
      write SetControl;
  end;

  TControlList = class(TOwnedCollection)
  private
    function  GetItem(Index : Integer) : TControlItem;
    procedure SetItem(Index : Integer; Value : TControlItem);
  protected
  public
    constructor Create(AOwner : TComponent);

    function Add : TControlItem;
    property Items[Index : Integer]       : TControlItem
      read GetItem
      write SetItem; default;
    published
  end;

  TDIBFeature = class(TComponent)
  private
    FControl        : TControl;
  protected
    property Control          : TControl
      read FControl;
  public
    class function CanApplyTo(aComponent : TPersistent) : Boolean; virtual;
    class function GetDisplayName : String; virtual;
    function GetOwner : TPersistent; override;
    procedure WndProc(var Message : TMessage; var Handled : Boolean); virtual; abstract;
  published
  end;

  TDIBFeatureItem = Class(TCollectionItem)
  private
    FSubPropertiesSize        : Integer;
    FSubProperties            : Pointer;
    FFeatureParameters        : String;
    FDIBFeature               : TDIBFeature;
    FEnabled                  : Boolean;
    FFeatureClassName         : String;
    procedure ReadParams(S : TStream);
    procedure SetFeatureClassName(const Value: String);
    procedure WriteParams(S : TStream);
  protected
    procedure DefineProperties(Filer : TFiler); override;
    procedure Loaded; virtual;
  public
    constructor Create(Collection : TCollection); override;
    Destructor Destroy; override;

    function GetDisplayName : String; override;
    procedure WndProc(var Message : TMessage; var Handled : Boolean); virtual;

    property DIBFeature       : TDIBFeature
      read FDIBFeature
      write FDIBFeature;

  published
    property Enabled          : Boolean
      read FEnabled
      write FEnabled
      default TRUE;
    property FeatureClassName : String
      read FFeatureClassName
      write SetFeatureClassName;
    property FeatureParameters : String
      read FFeatureParameters
      write FFeatureParameters;
  end;

  TDIBFeatures = class(TOwnedCollection)
  private
    FOwner    : TComponent;
    function  GetItem(Index : Integer) : TDIBFeatureItem;
    procedure SetItem(Index : Integer; Value : TDIBFeatureItem);
  protected
    procedure Loaded; virtual;
    procedure Update(Item : TCollectionItem); override;
  public
    constructor Create(AOwner : TComponent);

    function Add : TDIBFeatureItem;
    procedure WndProc(var Message : TMessage; var Handled : Boolean); virtual;

    property Owner            : TComponent
      read FOwner;
    property Items[Index : Integer]       : TDIBFeatureItem
      read GetItem
      write SetItem; default;
  published
  end;

  TDIBFeatureClass = class of TDIBFeature;

  TMoveableDIB = class(TDIBFeature)
  private
    FMoving         : Boolean;
    FOrigX,
    FOrigY,
    FX,
    FY              : Integer;
    FAllowVertical  : Boolean;
    FAllowHorizontal: Boolean;
    FBorderSize,
    FSnapSize       : Byte;

    procedure DoKeyDown(Message : TWMKey);
    procedure DoMouseDown(Message : TMessage);
    procedure DoMouseUp;
    procedure DoMouseMove(Message : TMessage);
  protected
  public
    constructor Create(AOwner : TComponent); override;
    class function GetDisplayName : String; override;
    procedure WndProc(var Message : TMessage; var Handled : Boolean); override;

  published
    property AllowHorizontal  : Boolean
      read FAllowHorizontal
      write FAllowHorizontal;
    property AllowVertical    : Boolean
      read FAllowVertical
      write FAllowVertical;
    property BorderSize       : Byte
      read FBorderSize
      write FBorderSize;
    property SnapSize         : Byte
      read FSnapSize
      write FSnapSize;
  end;

  THighlightDIB = class(TDIBFeature)
  private
    FOrigOpacity              : Byte;
    FHighlightOpacity         : Byte;
  protected
  public
    constructor Create(AOwner : TComponent); override;
    class function CanApplyTo(aComponent : TPersistent) : Boolean; override;
    class function GetDisplayName : String; override;
    procedure WndProc(var Message : TMessage; var Handled : Boolean); override;
  published
    property HighlightOpacity : Byte
      read FHighlightOpacity
      write FHighlightOpacity;
  end;

function ClassByName(Value : String) : TDIBFeatureClass;
procedure RegisterDIBFeature(aClass : TDIBFeatureClass);
var
  FeatureClasses    : Array of TDIBFeatureClass;

implementation

uses
  CDIBControl;

type
  EFeatureError = class(Exception);
  THackDIBControl = class(TCustomDIBControl);

function ClassByName(Value : String) : TDIBFeatureClass;
var
  X                 : Integer;
begin
  Result := nil;
  For X:=Length(FeatureClasses)-1 downto 0 do begin
    if CompareText(FeatureClasses[X].ClassName, Value)=0 then begin
      Result := FeatureClasses[X];
      Break;
    end;
  end;
end;

procedure RegisterDIBFeature(aClass : TDIBFeatureClass);
begin
  Classes.RegisterClass(aClass);
  Setlength(FeatureClasses,Length(FeatureClasses)+1);
  FeatureClasses[Length(FeatureClasses)-1] := aClass;
end;

{ TControlList }

function TControlList.Add: TControlItem;
begin
  Result := TControlItem(Inherited Add);
end;

constructor TControlList.Create(AOwner: TComponent);
begin
  inherited Create(AOwner, TControlItem);
end;



function TControlList.GetItem(Index: Integer): TControlItem;
begin
  Result := TControlItem(inherited GetItem(Index));
end;

procedure TControlList.SetItem(Index: Integer; Value: TControlItem);
begin
  inherited SetItem(Index, Value);
end;

{ TDIBFeatureItem }
constructor TDIBFeatureItem.Create(Collection: TCollection);
begin
  Inherited;
  FEnabled := True;
  FSubProperties := nil;
end;

procedure TDIBFeatureItem.DefineProperties(Filer: TFiler);
begin
  inherited;
  Filer.DefineBinaryProperty('DIBFeatureParameters',ReadParams,WriteParams, (FDIBFeature <> nil));
end;

destructor TDIBFeatureItem.Destroy;
begin
  if FSubProperties <> nil then Freemem(FSubProperties);
  if FDIBFeature <> nil then FDIBFeature.Free;
  inherited;
end;


function TDIBFeatureItem.GetDisplayName: String;
begin
  if FDIBFeature = nil then
    Result := 'DIB feature'
  else
    Result := FDIBFeature.GetDisplayName;
end;

procedure TDIBFeatureItem.ReadParams(S: TStream);
begin
  if S.Size > 0 then begin
    FSubPropertiesSize := S.Size;
    Getmem(FSubProperties,S.Size);
    S.Read(FSubProperties^, S.Size);
  end else
    FSubPropertiesSize := 0;
end;
(*
var
  Reader            : TReader;
begin
  Reader := TReader.Create(S, 4096);
  try
    Reader.IgnoreChildren := False;
    //This will create our DIBFeature item
    FeatureClassName := Reader.ReadString;
    Reader.ReadRootComponent(FDIBFeature);
  finally
    Reader.Free;
  end;
end;*)

procedure TDIBFeatureItem.Loaded;
var
  MS                : TMemoryStream;
  Reader            : TReader;
begin
  inherited;
  if FSubProperties <> nil then begin
    MS := TMemoryStream.Create;
    try
      MS.SetSize(FSubPropertiesSize);
      move(FSubProperties^, MS.Memory^, MS.Size);

      Reader := TReader.Create(MS, 4096);
      try
        //This will create our DIBFeature item
        Reader.IgnoreChildren := False;
        FeatureClassName := Reader.ReadString;
        Reader.ReadRootComponent(FDIBFeature);
      finally
        Reader.Free;
      end;
    finally
      MS.Free;
    end;
  end;
end;


procedure TDIBFeatureItem.SetFeatureClassName(const Value: String);
var
  TheClass          : TDIBFeatureClass;
begin
  TheClass := nil;
  if Value <> '' then begin
    TheClass := ClassByName(Value);
    if TheClass = nil then
      raise eFeatureError.Create(Value + ' has not been registered');
  end;

  if FDIBFeature <> nil then begin
    FDIBFeature.Free;
    FDIBFeature := nil;
  end;
  FFeatureClassName := Value;
  if TheClass <> nil then begin
    FDIBFeature := TheClass.Create(TControl(TDIBFeatures(Collection).GetOwner));
    FDIBFeature.FControl := TControl(TDIBFeatures(Collection).GetOwner);
  end;
end;

procedure TDIBFeatureItem.WndProc(var Message: TMessage;
  var Handled: Boolean);
begin
  if Enabled then
    if FDIBFeature <> nil then
      FDIBFeature.WndProc(Message, Handled);
end;

procedure TDIBFeatureItem.WriteParams(S: TStream);
var
  Writer            : TWriter;
begin
  Writer := TWriter.Create(S,4096);
  try
    Writer.IgnoreChildren := False;
    Writer.WriteString(FFeatureClassName);
    Writer.WriteRootComponent(FDIBFeature);
  finally
    Writer.Free;
  end;
end;

{ TDIBFeatures }

function TDIBFeatures.Add: TDIBFeatureItem;
begin
  Result := TDIBFeatureItem(inherited Add);
end;

constructor TDIBFeatures.Create(AOwner: TComponent);
begin
  inherited Create(AOwner, TDIBFeatureItem);
  FOwner := AOwner;
end;

function TDIBFeatures.GetItem(Index: Integer): TDIBFeatureItem;
begin
  Result := TDIBFeatureItem(inherited GetItem(Index));
end;

procedure TDIBFeatures.Loaded;
var
  X                 : Integer;
begin
  For X:=0 to Count-1 do Items[X].Loaded;
end;

procedure TDIBFeatures.SetItem(Index: Integer; Value: TDIBFeatureItem);
begin
  inherited SetItem(Index, Value);
end;

procedure TDIBFeatures.Update(Item: TCollectionItem);
begin
  inherited Update(Item);
end;

procedure TDIBFeatures.WndProc(var Message: TMessage;
  var Handled: Boolean);
var
  X                 : Integer;
begin
  for X:=0 to Count-1 do begin
    With Items[X] do
      WndProc(Message, Handled);
    if Handled then break;
  end;
end;

{ TMoveableDIB }

constructor TMoveableDIB.Create(AOwner : TComponent);
begin
  inherited;
  AllowVertical := True;
  AllowHorizontal := True;
  SnapSize := 1;
end;

procedure TMoveableDIB.DoKeyDown(Message: TWMKey);
begin
  with Message, Control do begin
    case CharCode of
      VK_UP : if AllowVertical then Top := Top - SnapSize;
      VK_DOWN : if AllowVertical then Top := Top + SnapSize;
      VK_Left: if AllowHorizontal then Left := Left - SnapSize;
      VK_RIGHT: if AllowHorizontal then Left := Left + SnapSize;
    end;
  end;
end;

procedure TMoveableDIB.DoMouseDown(Message: TMessage);
begin
  if Control = nil then exit;
  with TWMMouse(Message) do begin
    if ((FX >=BorderSize) or
        (FX <= Control.Width-BorderSize)) and
       ((FY >=BorderSize) or
        (FY <= Control.Height-BorderSize)) then
    begin
      FMoving := True;
      FX := XPos;
      FY := YPos;
    end;
  end;
  with Control do begin
    FOrigX := Left;
    FOrigY := Top;
  end;
end;

procedure TMoveableDIB.DoMouseMove(Message: TMessage);
var
  DX,
  DY                          : Integer;
begin
  if FMoving then with TWMMouse(Message) do begin
    if AllowHorizontal then
      DX := (XPos - FX)
    else
      DX := 0;
    if AllowVertical then
      DY := (YPos - FY)
    else
      DY := 0;
    if SnapSize > 1 then begin
      DX := DX div SnapSize * SnapSize;
      DY := DY div SnapSize * SnapSize;
    end;
    if Control <> nil then with Control do begin
      SetBounds(Left+DX, Top+DY, Width,Height);
    end;
  end;
end;

procedure TMoveableDIB.DoMouseUp;
begin
  FMoving := False;
end;

class function TMoveableDIB.GetDisplayName: String;
begin
  Result := 'Moveable DIB';
end;

procedure TMoveableDIB.WndProc(var Message: TMessage;
  var Handled: Boolean);
begin
  case Message.msg of
    WM_LButtonDown  : DoMouseDown(message);
    WM_LButtonUp    : DoMouseUp;
    WM_KeyDown      : DoKeyDown(TWMKey(Message));
    WM_MouseMove    : DoMouseMove(message)
  end;
end;
{ TDIBFeature }

class function TDIBFeature.CanApplyTo(aComponent: TPersistent): Boolean;
begin
  Result := True;
end;

class function TDIBFeature.GetDisplayName: String;
begin
  Result := 'Unknown feature';
end;

function TDIBFeature.GetOwner: TPersistent;
begin
  Result := FControl;
end;

{ THighlightDIB }

class function THighlightDIB.CanApplyTo(aComponent: TPersistent): Boolean;
begin
  if not (aComponent is TCustomDIBControl) then
    Result := False
  else
    Result := inherited CanApplyTo(aComponent);
end;

constructor THighlightDIB.Create(AOwner: TComponent);
begin
  inherited;
  FHighlightOpacity := 255;
end;

class function THighlightDIB.GetDisplayName: String;
begin
  Result := 'Highlight dib';
end;

procedure THighlightDIB.WndProc(var Message: TMessage;
  var Handled: Boolean);
begin
  if Control is TCustomDIBControl then with THackDIBControl(Control) do begin
    case Message.Msg of
      WM_SetFocus: if not Focused and not MouseInControl then
        begin
          FOrigOpacity := Opacity;
          Opacity := HighlightOpacity;
        end;

      WM_KillFocus: if Focused and not MouseInControl then
        begin
          Opacity := FOrigOpacity;
        end;

      CM_MouseEnter : if not Focused then
        begin
          if MouseCapture then exit;
          FOrigOpacity := Opacity;
          Opacity := HighlightOpacity;
        end;

      WM_LButtonUp :
        begin
          MouseCapture := False;
        end;

      CM_MouseLeave : if not Focused then
        begin
          if MouseCapture then exit;
          Opacity := FOrigOpacity;
        end;
    end;
  end;
end;


{ TControlItem }

procedure TControlItem.AssignTo(Dest: TPersistent);
begin
  if Dest is TControlItem then
    TControlItem(Dest).Control := FControl
  else
    inherited;
end;

function TControlItem.GetDisplayName: String;
begin
  if Control = nil then
    Result := inherited GetDisplayName
  else
    if Control.Name <> '' then
      Result := Control.Name
    else
      if Control.ClassName <> '' then
        Result := Control.ClassName
      else
        Result := inherited GetDisplayName;
end;

procedure TControlItem.SetControl(const Value: TControl);
var
  X                           : Integer;
begin
  if Value = nil then
    raise EFeatureError.Create('You cannot set Control to nil');
  for X:=0 to Collection.Count-1 do
    if (TControlList(Collection).Items[X].Control = Value) and
       (Collection.Items[X] <> Self) then
      raise EFeatureError.Create('Control already exists in list.');
  FControl := Value;
end;

initialization
  RegisterDIBFeature(TMoveableDIB);
  RegisterDIBFeature(THighlightDIB);

end.
